'''Unittests for Swarm class.'''

import os
import glob
import pickle
import unittest
from shutil import rmtree

import pandas as pd
from sklearn.model_selection import train_test_split

import ensembleset.dataset as ds
from ensembleswarm.swarm import Swarm

class TestSwarm(unittest.TestCase):
    '''Tests for ensemble swarm class.'''

    def setUp(self):
        '''Dummy swarm instance for tests.'''

        # Clear data directory
        if os.path.isdir('data'):
            rmtree('data')

        # Ensembleset parameters
        self.n_datasets = 3
        self.n_features = 2
        self.n_steps = 3

        # Load and prep calorie data for testing
        data_df=pd.read_csv('tests/calories.csv')
        data_df=data_df.sample(frac=0.01)
        data_df.drop('id', axis=1, inplace=True, errors='ignore')
        train_df, test_df=train_test_split(data_df, test_size=0.5)
        train_df.reset_index(inplace=True, drop=True)
        test_df.reset_index(inplace=True, drop=True)

        # Set-up ensembleset
        self.dataset = ds.DataSet(
            label='Calories',
            train_data=train_df,
            test_data=test_df,
            string_features=['Sex']
        )

        # Generate datasets
        self.dataset.make_datasets(
            n_datasets=self.n_datasets,
            n_features=self.n_features,
            n_steps=self.n_steps
        )

        # Initialize ensembleswarm
        self.swarm = Swarm()


    def test_class_arguments(self):
        '''Tests assignments of class attributes from user arguments.'''

        self.assertTrue(isinstance(self.swarm.ensembleset, str))
        self.assertTrue(isinstance(self.swarm.models, dict))

        with self.assertRaises(TypeError):
            _ = Swarm(ensembleset=0.0)


    def test_train_swarm(self):
        '''Tests fitting of ensemble swarm'''

        self.swarm.train_swarm()

        self.assertTrue(os.path.isdir('data/swarm'))

        swarms=glob.glob('data/swarm/*pkl')
        self.assertEqual(len(swarms), self.n_datasets)

        with open(swarms[1], 'rb') as input_file:
            swarm = pickle.load(input_file)

        self.assertTrue(isinstance(swarm, dict))



'''Creates and trains a swarm of level II regression ensembles.'''

import pickle
import copy
from pathlib import Path

import h5py
import numpy as np

import ensembleswarm.regressors as regressors
class Swarm:
    '''Class to hold ensemble model swarm.'''

    def __init__(
            self,
            ensembleset: str = 'data/dataset.h5'
        ):

        # Check user argument types
        type_check = self.check_argument_types(
            ensembleset
        )

        # If the type check passed, assign arguments to attributes
        if type_check is True:
            self.ensembleset = ensembleset

        self.models = regressors.MODELS


    def check_argument_types(self,
            ensembleset: str
    ) -> bool:

        '''Checks user argument types, returns true or false for all passing.'''

        check_pass = False

        if isinstance(ensembleset, str):
            check_pass = True

        else:
            raise TypeError('Ensembleset path is not a string.')

        return check_pass

    def train_swarm(self) -> None:
        '''Trains an instance of each regressor type on each member of the ensembleset.'''

        Path('data/swarm').mkdir(parents=True, exist_ok=True)

        with h5py.File(self.ensembleset, 'r') as hdf:

            num_datasets=len(list(hdf['train'].keys())) - 1

            for i in range(num_datasets):

                print(f'\nBuilding swarm {i+1} of {num_datasets}')
                models=copy.deepcopy(self.models)

                for model_name in models.keys():

                    print(f' Fitting {model_name}')

                    _=models[model_name].fit(
                        np.array(hdf[f'train/{i}']), 
                        np.array(hdf['train/labels'])
                    )

                with open(f'data/swarm/{i}.pkl', 'wb') as output_file:
                    pickle.dump(models, output_file)




'''Regressor definitions and hyperparameter distributions for GridSearchCV with SciKit-learn.'''

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import QuantileRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import LinearSVR, SVR
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import AdaBoostRegressor
from sklearn.linear_model import SGDRegressor
from xgboost import XGBRegressor
from catboost import CatBoostRegressor
from lightgbm import LGBMRegressor

REGRESSOR_JOBS = -1

MODELS={
    'Linear regression': LinearRegression(),
    'Quantile regression': QuantileRegressor(),
    'Nearest Neighbors':KNeighborsRegressor(n_jobs=REGRESSOR_JOBS),
    'Linear SVM':LinearSVR(max_iter=1000),
    'RBF SVM':SVR(kernel='rbf'),
    'Polynomial SVM':SVR(kernel='poly'),
    'Gaussian Process':GaussianProcessRegressor(),
    'Decision Tree':DecisionTreeRegressor(),
    'Random Forest':RandomForestRegressor(n_jobs=REGRESSOR_JOBS),
    'Neural Net':MLPRegressor(),
    'AdaBoost':AdaBoostRegressor(),
    'SGD':SGDRegressor(penalty='elasticnet'),
    'XGBoost':XGBRegressor(n_jobs=REGRESSOR_JOBS),
    'CatBoost':CatBoostRegressor(thread_count=REGRESSOR_JOBS),
    'LightGBM':LGBMRegressor(n_jobs=REGRESSOR_JOBS)
}

HYPERPARAMETERS={
    'Linear regression':{
        'fit_intercept':[True,False]
    },
    'Quantile regression':{
        'fit_intercept':[True,False],
        'alpha':[1.0]
    },
    'Nearest Neighbors':{
        'n_neighbors': [3, 4, 5],
        'weights': ['distance'],
        'algorithm': ['ball_tree'],
        'leaf_size': [30, 40, 50]
    },
    'Linear SVM':{
        'C': [1.0], 
        'epsilon': [0.0],
    },
    'RBF SVM':{
        'C': [0.24, 0.25, 0.26],
        'epsilon': [0.01, 0.1, 1]
    },
    'Polynomial SVM':{
        'C': [0.0125, 0.025, 0.05],
        'epsilon': [0.01, 0.1, 1],
        'degree': [2, 3],
        'coef0': [0.5, 1.0, 2]
    },
    'Gaussian Process':{
        'n_restarts_optimizer': [0]
    },
    'Decision Tree':{
        'criterion': ['squared_error'],
        'splitter': ['best'],
        'max_depth': [None, 100, 500],
        'max_features': [None, 0.5, 0.9]
    },
    'Random Forest':{
        'n_estimators': [300, 400, 500],
        'criterion': ['squared_error'],
        'max_depth': [None, 6, 7, 8],
        'max_features': [None, 0.75, 0.80, 0.85],
        'ccp_alpha': [0.0]
    },
    'Neural Net':{
        'hidden_layer_sizes': [3, 6, 12],
        'alpha': [0.2, 0.4, 0.8],
        'learning_rate': ['adaptive'],
        'max_iter': [1000]
    },
    'AdaBoost':{
        'n_estimators': [300, 400, 500],
        'learning_rate': [0.01, 0.02, 0.03],
        'loss': ['linear']
    },
    'SGD':{
        'loss': ['squared_error'],
        'alpha': [0.004, 0.005, 0.006],
        'l1_ratio': [0.55, 0.6, 0.65],
        'learning_rate': ['invscaling']
    },
    'XGBoost':{
        'n_estimators': [15, 20, 25],
        'max_depth': [2, 3, 4],
        'subsample': [1.0]
    },
    'CatBoost':{
        'n_estimators': [200, 300, 400],
        'depth': [1, 2, 3],
        'model_size_reg':[1e-9, 1e-8, 1e-7]
    },
    'LightGBM':{
        'learning_rate': [0.09, 0.1, 0.11],
        'n_estimators': [75, 100, 125],
        'max_depth': [1, 2, 3],
        'subsample': [0.25, 0.3, 0.35]
    }
}
